/*
 * ast.c: abstract syntax tree for datadec
 *
 *      This module provides the internal structures needed
 *      for the data declaration builder.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bool.h"
#include "ast.h"

#define NEW(t)	        ((t) malloc( sizeof(struct t) ))
#define COPYOF(new,old) {new=malloc(1+strlen(old));if(new)strcpy(new,old);}

paramlist build_paramlist( char *type, char *id, paramlist next )
{
        paramlist new = NEW(paramlist);
        COPYOF(new->type, type);
        COPYOF(new->name, id);
        new->next  = next;
        return new;
}

void print_paramlist( FILE *out, paramlist p )
{
        while( p != NULL )
	{
                fprintf(out, "%s %s", p->type, p->name);
		p = p->next;
                if( p != NULL )
		{
			fputs( ", ", out );
                }
        }
}

shapelist build_shapelist( char *id, paramlist p, shapelist next )
{
        shapelist new = NEW(shapelist);
        COPYOF(new->name, id);
        new->params = p;
        new->next   = next;
        return new;
}

void print_shapelist( FILE *out, shapelist s )
{
        while( s != NULL )
	{
		fputs( s->name, out );
		if (s->params)
		{
			fputs( "( ", out );
			print_paramlist( out, s->params );
			fputs( " )", out );
		}
		putc(' ', out );
                s = s->next;
		if( s != NULL )
		{
                        fputs( "or ", out );
                }
        }
}

declnlist build_declnlist( char *name, shapelist s, declnlist next )
{
        declnlist new = NEW(declnlist);
        COPYOF(new->name, name);
        new->shapes = s;
        new->next   = next;
        return new;
}

void print_declnlist( FILE *out, declnlist d )
{
        for ( ; d; d = d->next)
	{
                fputs(d->name, out);
		fputs(" = ", out);
                print_shapelist(out, d->shapes);
                fputs(";\n", out);
        }
}
