#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bool.h"
#include "ast.h"
#include "decs.h"

static void h_declns( declnlist d, char *base );
static void data_decls( declnlist decs );
static void declare_type( decln d );
static void protos( declnlist d );
static void proto_types( decln d );
static void cons_fn_proto( char *dname, shape s );
static void deconskind_proto( char *dname );
static void decons_fn_proto( char *dname, shape s );
static void print_fn_proto( char * name );

static int numtabs = 0;
static FILE *outfile;

#define indent() numtabs++
#define outdent() numtabs--
#define usefile(f)	outfile = f, numtabs = 0

/*VARARGS*/
static void line( fmt, a, b, c, d ) char *fmt; long a, b, c, d;
//static void line( char *fmt, long a, long b, long c, long d )
{
	int i;
	for( i=numtabs; i; i-- ) fputc( '\t', outfile );
	fprintf( outfile, fmt, a, b, c, d );
	fputc( '\n', outfile );
}

void make_declns( declnlist d, char *base )
{
	printf( "Making data declarations in %s.h, %s.c file coming soon\n",
		base, base );
	h_declns( d, base );
	/* Coming soon: c_declns( d, base ); */
} 

static void h_declns( declnlist d, char *base )
{
	char	tempname[256];
	FILE	*hfile;
	sprintf( tempname, "%s.h", base );
	hfile = fopen( tempname, "w" );
	if( hfile == NULL )
	{
		fprintf( stderr, "minidatadec: can't create '%s'\n", tempname );
		exit(1);
	}
	usefile( hfile );
	line( "/*" );
	line( " * Automatically Generated by MiniDataDec" );
	line( " */\n" );
	line( "typedef char *string;\n" );
	line( "typedef char BOOL;" );
	line( "#define TRUE 1" );
	line( "#define FALSE 0\n" );
	line( "#define NEW(t) ((t)malloc(sizeof(struct t)))\n\n" );
	data_decls( d );
	protos( d );
	fclose( hfile );
}

/* ------------------------ Declare the types ------------------------------ */

static void data_decls( declnlist decs )
{
	declnlist	d;
	for( d=decs; d != NULL; d=d->next )
	{
		line("struct %s;", d->name);
		line("typedef struct %s *%s;", d->name, d->name );
	}
	fputc( '\n', outfile );
	fputc( '\n', outfile );
	for( d=decs; d != NULL; d=d->next )
	{
		declare_type( d );
	}
}

static void declare_type( decln d )
{
	char *dname = d->name;
	line( "/* ---- Type %s ---- */\n", dname );
	line( "typedef enum {" );
	indent();
	shapelist	s;
	for( s = d->shapes; s != NULL; s=s->next )
	{
		line( "%s_is_%s,", dname, s->name );
	}
	outdent();
	line( "} kind_of_%s;\n", dname );
	line( "struct %s {", dname );
	indent();
	line( "kind_of_%s\ttag;", dname );
	line( "union {" );
	indent();
	for( s = d->shapes; s != NULL; s=s->next )
	{
		if( s->params == NULL ) continue;
		/* internal struct */
		line( "struct {" );
		indent();
		paramlist p = s->params;
		for( ; p != NULL; p=p->next )
		{
			line( "%s\t%s;", p->type, p->name );
		}
		outdent();
		line( "} %s;", s->name );
	}
	outdent();
	line( "} u;" );
	outdent();
	line( "};\n" );
}

/* ---------------------- Declare all prototypes -------------------------- */

static void protos( declnlist d )
{
	line( "\n/* Prototypes for all types */\n" );
	for( ; d != NULL; d = d->next )
	{
		proto_types( d );
	}
}

static void proto_types( decln d )
{
	char *dname = d->name;
	shapelist s = d->shapes;
	/* constructor function prototypes */
	for( ; s != NULL; s = s->next )
	{
		cons_fn_proto( dname, s );
	}
	/* deconstructor kind prototype */
	deconskind_proto( dname );
	/* deconstructor function prototypes */
	for( s = d->shapes; s != NULL; s = s->next )
	{
		if( s->params != NULL )
		{
			decons_fn_proto( dname, s );
		}
	}
	print_fn_proto( dname );
}

/* ------------------------------- Constructors --------------------------- */

static void cons_fn_proto( char *dname, shape s )
{
	fprintf( outfile, "extern %s %s_%s( ", dname, dname, s->name );
	if( s->params == NULL )
	{
		fputs( "void ", outfile );
	} else
	{
		BOOL first = TRUE;
		paramlist p;
		for( p = s->params; p != NULL ; p=p->next, first=FALSE )
		{
			if( !first ) fputs( ", ", outfile );
			fputs( p->type, outfile );
			fputc( ' ', outfile );
		}
	}
	fputs( ");\n", outfile );
}

/* ------------------------- Deconstructors ------------------------------- */

/* ------------------- Deconstructor Kind prototype ----------------------- */
static void deconskind_proto( char *dname )
{
	fprintf( outfile, "extern kind_of_%s %s_kind( %s );\n",
		dname, dname, dname );
}

/* --------------------- Deconstructor prototypes ------------------------ */
static void decons_fn_proto( char *dname, shape s )
{
	if( s->params == NULL )
	{
		fprintf( stderr, "decons_fn_proto: no fields in shape!!\n" );
		exit(1);
	}
	fprintf( outfile, "extern void get_%s_%s( %s ",
		dname, s->name, dname );
	paramlist p;
	for( p = s->params; p != NULL ; p=p->next )
	{
		fprintf( outfile, ", %s * ", p->type );
	}
	fputs( ");\n", outfile );
}

/* -------------------------- Print functions ----------------------------- */
static void print_fn_proto( char *name )
{
	fprintf( outfile, "extern void print_%s( FILE *, %s );\n", name, name );
}
