#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bool.h"
#include "ast.h"
#include "decs.h"

static void h_declns( declnlist d, char *base );
static void c_declns( declnlist d, char *base );
static void data_decls( declnlist decs );
static void declare_type( decln d );
static void protos( declnlist d );
static void proto_types( decln d );
static void cons_fns( declnlist d );
static void cons_fn_proto( char *dname, shape s );
static void cons_fn( char *dname, shape s );
static void decons_fns( declnlist d );
static void decons_fns_type( decln d );
static void deconskind_proto( char *dname );
static void deconskind_fn( char *dname );
static void decons_fn_proto( char *dname, shape s );
static void decons_fn( char *dname, shape s );
static void print_fn_proto( char * name );
static void print_fns( declnlist d );
static void print_fn_shape( char *dname, shape s );
static void print_all_params( shape s );
static void print_param( char *sname, param p );

static int numtabs = 0;
static FILE *outfile;

#define indent() numtabs++
#define outdent() numtabs--
#define usefile(f)	outfile = f, numtabs = 0

/*VARARGS*/
static void line( fmt, a, b, c, d ) char *fmt; long a, b, c, d;
//static void line( char *fmt, long a, long b, long c, long d )
{
	int i;
	for( i=numtabs; i; i-- ) fputc( '\t', outfile );
	fprintf( outfile, fmt, a, b, c, d );
	fputc( '\n', outfile );
}

void make_declns( declnlist d, char *base )
{
	printf( "datadec: Making data declarations in %s.[ch]\n", base );
	h_declns( d, base );
	c_declns( d, base );
}

static void h_declns( declnlist d, char *base )
{
	char	tempname[256];
	FILE	*hfile;
	sprintf( tempname, "%s.h", base );
	hfile = fopen( tempname, "w" );
	if( hfile == NULL )
	{
		fprintf( stderr, "datadec: can't create '%s'\n", tempname );
		exit(1);
	}
	usefile( hfile );
	line( "/*" );
	line( " * Automatically Generated by MiniDataDec" );
	line( " */\n" );
	line( "typedef char *string;\n" );
	line( "typedef char BOOL;" );
	line( "#define TRUE 1" );
	line( "#define FALSE 0\n" );
	line( "#define NEW(t) ((t)malloc(sizeof(struct t)))\n\n" );
	data_decls( d );
	protos( d );
	fclose( hfile );
}

static void c_declns( declnlist d, char *base )
{
	char	tempname[256];
	FILE	*cfile;
	sprintf( tempname, "%s.c", base );
	cfile = fopen( tempname, "w" );
	if( cfile == NULL )
	{
		fprintf( stderr, "datadec: can't create '%s'\n", tempname );
		exit(1);
	}
	usefile( cfile );
	line( "/*" );
	line( " * Automatically Generated by MiniDataDec" );
	line( " */\n" );
	line( "#include <stdio.h>" );
	line( "#include <stdlib.h>" );
	line( "#include \"%s.h\"\n\n", base );
	cons_fns( d );
	decons_fns( d );
	print_fns( d );
	fclose( cfile );
}

/* ------------------------ Declare the types ------------------------------ */

static void data_decls( declnlist decs )
{
	declnlist	d;
	for( d=decs; d != NULL; d=d->next )
	{
		line("struct %s;", d->name);
		line("typedef struct %s *%s;", d->name, d->name );
	}
	fputc( '\n', outfile );
	fputc( '\n', outfile );
	for( d=decs; d != NULL; d=d->next )
	{
		declare_type( d );
	}
}

static void declare_type( decln d )
{
	char *dname = d->name;
	line( "/* ---- Type %s ---- */\n", dname );
	line( "typedef enum {" );
	indent();
	shapelist	s;
	for( s = d->shapes; s != NULL; s=s->next )
	{
		line( "%s_is_%s,", dname, s->name );
	}
	outdent();
	line( "} kind_of_%s;\n", dname );
	line( "struct %s {", dname );
	indent();
	line( "kind_of_%s\ttag;", dname );
	line( "union {" );
	indent();
	for( s = d->shapes; s != NULL; s=s->next )
	{
		if( s->params == NULL ) continue;
		/* internal struct */
		line( "struct {" );
		indent();
		paramlist p = s->params;
		for( ; p != NULL; p=p->next )
		{
			line( "%s\t%s;", p->type, p->name );
		}
		outdent();
		line( "} %s;", s->name );
	}
	outdent();
	line( "} u;" );
	outdent();
	line( "};\n" );
}

/* ---------------------- Declare all prototypes -------------------------- */

static void protos( declnlist d )
{
	line( "\n/* Prototypes for all types */\n" );
	for( ; d != NULL; d = d->next )
	{
		proto_types( d );
	}
}

static void proto_types( decln d )
{
	char *dname = d->name;
	shapelist s = d->shapes;
	/* constructor function prototypes */
	for( ; s != NULL; s = s->next )
	{
		cons_fn_proto( dname, s );
	}
	/* deconstructor kind prototype */
	deconskind_proto( dname );
	/* deconstructor function prototypes */
	for( s = d->shapes; s != NULL; s = s->next )
	{
		if( s->params != NULL )
		{
			decons_fn_proto( dname, s );
		}
	}
	print_fn_proto( dname );
}

/* ------------------------------- Constructors --------------------------- */

static void cons_fns( declnlist d )
{
	for( ; d != NULL; d=d->next )
	{
		shapelist s = d->shapes;
		for( ; s != NULL; s=s->next )
		{
			cons_fn( d->name, s );
		}
	}
}

static void cons_fn_proto( char *dname, shape s )
{
	fprintf( outfile, "extern %s %s_%s( ", dname, dname, s->name );
	if( s->params == NULL )
	{
		fputs( "void ", outfile );
	} else
	{
		BOOL first = TRUE;
		paramlist p;
		for( p = s->params; p != NULL ; p=p->next, first=FALSE )
		{
			if( !first ) fputs( ", ", outfile );
			fputs( p->type, outfile );
			fputc( ' ', outfile );
		}
	}
	fputs( ");\n", outfile );
}

static void cons_fn( char *dname, shape s )
{
	fprintf( outfile, "%s %s_%s( ", dname, dname, s->name );
	BOOL first = TRUE;
	if( s->params == NULL )
	{
		fprintf( outfile, "void" );
	}
	paramlist p;
	for( p=s->params; p != NULL; p=p->next, first = FALSE )
	{
		if( !first ) fputs( ", ", outfile );
		fprintf( outfile, "%s %s", p->type, p->name );
	}
	fputs( " )", outfile );
	line( "\n{" );
	indent();
	line( "%s\tnew = NEW(%s);", dname, dname );
	line( "new->tag = %s_is_%s;", dname, s->name );
	for( p=s->params; p != NULL; p=p->next )
	{
		/* an internal struct */
		line( "new->u.%s.%s = %s;", s->name, p->name, p->name );
	}
	line( "return new;" );
	outdent();
	line( "}\n\n" );
}

/* ------------------------- Deconstructors ------------------------------- */

static void decons_fns( declnlist d )
{
	for( ; d != NULL; d=d->next )
	{
		decons_fns_type( d );
	}
}

static void decons_fns_type( decln d )
{
	/* We need a deconstructor kind function */
	deconskind_fn( d->name );

	/* We need deconstructors for shapes with params */
	shapelist s;
	for( s = d->shapes; s != NULL; s=s->next )
	{
		if( s->params != NULL )
		{
			decons_fn( d->name, s );
		}
	}
}

/* ------------------- Deconstructor Kind procedure ----------------------- */
static void deconskind_proto( char *dname )
{
	fprintf( outfile, "extern kind_of_%s %s_kind( %s );\n",
		dname, dname, dname );
}

static void deconskind_fn( char *dname )
{
	fprintf( outfile, "kind_of_%s %s_kind( %s this )\n",
		 dname, dname, dname );
	line( "{" );
	indent();
	line( "return this->tag;" );
	outdent();
	line( "}\n\n" );
}

/* ----------------------- Deconstructor functions ------------------------ */
static void decons_fn_proto( char *dname, shape s )
{
	if( s->params == NULL )
	{
		fprintf( stderr, "decons_fn_proto: no fields in shape!!\n" );
		exit(1);
	}
	fprintf( outfile, "extern void get_%s_%s( %s ",
		dname, s->name, dname );
	paramlist p;
	for( p = s->params; p != NULL ; p=p->next )
	{
		fprintf( outfile, ", %s * ", p->type );
	}
	fputs( ");\n", outfile );
}

static void decons_fn( char *dname, shape s )
{
	if( s->params == NULL )
	{
		fprintf( stderr, "decons_fn: no fields in shape!!\n" );
		exit(1);
	}
	fprintf( outfile, "void get_%s_%s( %s this", dname, s->name, dname );
	paramlist p;
	for( p=s->params; p != NULL; p=p->next )
	{
		fprintf( outfile, ", %s *%s", p->type, p->name );
	}
	fprintf( outfile, " )" );
	line( "\n{" );
	indent();
	for( p=s->params; p != NULL; p=p->next )
	{
		/* internal struct */
		line( "*%s = this->u.%s.%s;",
			p->name, s->name, p->name );
	}
	outdent();
	line( "}\n\n" );
}

/* -------------------------- Print functions ----------------------------- */

static void print_fn_proto( char *name )
{
	fprintf( outfile, "extern void print_%s( FILE *, %s );\n", name, name );
}

static void print_fns( declnlist d )
{
	for( ; d != NULL; d = d->next )
	{
		line( "void print_%s( FILE *f, %s p )",d->name, d->name );
		line( "{" );
		indent();
		line( "switch( p->tag )" );
		line( "{" );
		shapelist s;
		for( s = d->shapes; s != NULL; s = s->next )
		{
			print_fn_shape( d->name, s );
		}
		line( "default:" );
		indent();
		line( "fprintf( stderr," );
		line( " \"print_%s: impossible tag %%d\\n\", p->tag );",
			d->name );
		line( "exit(1);" );
		outdent();
		line( "}" );
		outdent();
		line( "}\n\n" );
	}
}

static void print_fn_shape( char *dname, shape s )
{
	line( "case %s_is_%s:", dname, s->name );
	indent();
	print_all_params( s );
	line( "break;" );
	outdent();
}

static void print_all_params( shape s )
{
	paramlist	p;
	line( "fputs( \"%s\", f );", s->name );
	if( s->params )
	{
		line( "fputc( '(', f );" );
	}
	for( p = s->params; p != NULL; p = p->next )
	{
		print_param( s->name, p );
		if( p->next != NULL )
		{
			line( "fputc( ',', f );" );
		}
	}
	if( s->params )
	{
		line( "fputc( ')', f );" );
	}
}

static void print_param( char *sname, param p )
{
	char pname[200];
	sprintf( pname, "p->u.%s.%s", sname, p->name );
	if( streq( p->type, "int" ) )
	{
		line( "fprintf( f, \"%%d\", %s );", pname );
	} else if( streq( p->type, "char" ) )
	{
		line( "fputc( %s, f );", pname );
	} else if( streq( p->type, "BOOL" ) )
	{
		line( "fputs( %s?\"true\":\"false\", f );", pname );
	} else if( streq( p->type, "string" ) )
	{
		line( "fputs( %s, f );", pname );
	} else
	{
		line( "print_%s( f, %s );", p->type, pname );
	}
}
